# $Id: pybscan.py,v 1.4 2009/08/19 08:05:04 jhh Exp $
#
# Author: Jaap-Henk Hoepman (jhh@xs4all.nl)
#
# Functions to scan BiBTeX files
#
# (for python3)

import os,string

# report errors by throwing this exception
class BibError(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)


class BibScanner():

    # character buffer that contains a single bibfile as a single string
    charbuf = ""

    # pointer (in charbuf) to the current character to be read
    cur = 0

    # line number corresponding to the current character to be read
    # (used for error reporting)
    lineno = 0

    # raise an error, recording the current line number in the file
    def error(m):
        line = "Line " + str(lineno)
        raise BibError(line + ": " + m)

    # Create a bibscanner: open and load the specified bibtex file and
    # initialise the scanner (if is_file is true) or
    # use name itself as string to be scanned otherwise
    def __init__(self,name,is_file):
        if is_file:
            # add .bib to the filename if not present
            bibname = os.path.splitext (name) [0] + '.bib'
            # make sure \r\n is returned as \n
            # also open in the right encoding; bib files may contain umlauts etc
            f = open(bibname,"r",encoding='latin-1', newline=None)
            # read the whole file
            self.charbuf = f.read()
            f.close()
        else:
            self.charbuf = name
        self.cur = 0
        self.lineno = 1

    # return the character at position i; return None if index out of range
    def atch(self,i):
        if i < 0:
            return None
        elif i >= len(self.charbuf):
            return None
        else:
            return self.charbuf[i]

    # return the character before the current one, None if at beginning of file
    def prevch(self):
        return(self.atch(self.cur-1))

    # return the current character, None if at EOF
    def curch(self):
        return(self.atch(self.cur))        

    # advance the charcter pointer; update the line number where appropriate
    # (does nothing if at EOF)
    def nextch(self):
        if self.curch() == "\n":
            self.lineno += 1
        if self.cur < len(self.charbuf):            
            self.cur += 1    

    # skip until the end of line.
    # when this returns, curch() returns \n, or None if EOF
    def skip_until_nl(self):
        while (self.curch() != None) and (self.curch() != '\n'):
            self.nextch()

    # skip whitespace; skip all characters in string.whitespace
    # and skip over comments that start with "%" and are terminated by a newline
    # curch() returns the first nonwhitespace char, or None if at EOF
    def skip_whitespace(self):
        while ((self.curch() != None) and
               ((self.curch() == "%") or (self.curch() in string.whitespace))):
            if self.curch() == '%':
                self.skip_until_nl()
            self.nextch()

    # skip initial part of bib file upto first line starting with @; curch points
    # to this after call (or None if EOF reached)
    def skip_until_at_line(self):
        while ((self.curch() != None) and
               ( (self.curch() != "@") or
                 ( (self.prevch() != "\n") and (self.prevch() != None ) ))):
            self.nextch()

    # skip whitespace and check if the next char on the input is in the set s
    # if so, return that character, and advance the character pointer.
    # If not or at EOF, return None
    def expect(self,s):
        self.skip_whitespace()
        c = self.curch()
        if (c == None) or (c not in s):
            return None
        else:
            self.nextch()
            return c

    # skip whitespace and check if the next char on the input is in s
    # if so, return it, and advance the character pointer.
    # If not or at EOF, report an error.
    def required(self,s):
        c = self.expect(s)
        if c == None:
            self.error("Expecting " + s + " and seeing " + self.curch())
        else:
            return c

